from typing import List

import h5py

from funlbm.base import Worker
from funlbm.particle import Particle, ParticleConfig
from funlbm.util import logger

from .base import create_particle


class ParticleSwarm(Worker):
    def __init__(self, configs: List[ParticleConfig] = [], device="cpu"):
        self.particles: List[Particle] = []
        for config in configs:
            self.particles.append(create_particle(config, device=device))

    def init(self, *args, **kwargs):
        for particle in self.particles:
            particle.init(*args, **kwargs)
        self.save_particle()

    def update(self, *args, **kwargs):
        for particle in self.particles:
            particle.update(*args, **kwargs)

    def save_particle(self):
        with h5py.File("./funlbm-particle-swarm-lagrange.h5", "w") as fw:
            for i, particle in enumerate(self.particles):
                fw.create_dataset(f"particle{i}", data=particle._lagrange.cpu().numpy())
        logger.success("save particle lagrange success.")

    def dump_checkpoint(self, group: h5py.Group = None, *args, **kwargs):
        for i, particle in enumerate(self.particles):
            sub_group = group.create_group(f"particle_{str(i).zfill(6)}")
            particle.dump_checkpoint(sub_group, *args, **kwargs)

    def load_checkpoint(self, group: h5py.Group = None, *args, **kwargs):
        for i, particle in enumerate(self.particles):
            sub_group = group.get(f"particle_{str(i).zfill(6)}")
            particle.load_checkpoint(sub_group, *args, **kwargs)


def create_particle_swarm(
    configs: List[ParticleConfig] = [], device="cpu", *args, **kwargs
):
    return ParticleSwarm(configs=configs, device=device, *args, **kwargs)



from .base import Ellipsoid, Particle, ParticleConfig, Sphere, create_particle
from .swarm import ParticleSwarm, create_particle_swarm

__all__ = [
    "Particle",
    "Ellipsoid",
    "Sphere",
    "ParticleConfig",
    "ParticleSwarm",
    "create_particle_swarm",
    "create_particle",
]



import numpy as np
import torch
from funutil import run_timer
from funutil.cache import cache

from funlbm.config import Boundary, BoundaryCondition
from funlbm.flow import Flow, FlowConfig


def cul_u(y, z, a, b, size=100):
    # 预计算常量以提高性能
    n = np.arange(size)
    pi = np.pi
    n_terms = 2 * n + 1

    # 分解计算步骤提高可读性
    base_term = 32 * (-1) ** (n + 1) / (n_terms**3 * pi**3)
    y_term = np.cos(n_terms * pi * y / (2 * a))
    z_term = np.cosh(n_terms * pi * z / (2 * a)) / np.cosh(n_terms * pi * b / (2 * a))

    # 计算级数和
    series_sum = np.sum(base_term * y_term * z_term)
    result = 1 - (y / a) ** 2 + series_sum

    return round(result, 6)


@cache
def init_u(a, b, u_max=0.01, n_max=100):
    # 使用 meshgrid 创建坐标网格
    y = np.linspace(-(a - 1) / 2, (a - 1) / 2, a)
    z = np.linspace(-(b - 1) / 2, (b - 1) / 2, b)
    Y, Z = np.meshgrid(y, z, indexing="ij")

    # 向量化计算
    res = u_max * np.vectorize(lambda y, z: cul_u(y, z, a / 2.0, b / 2.0, n_max))(Y, Z)
    return torch.from_numpy(res)


def tran3d(direction, total):
    """
    计算流场边界索引
    Args:
        direction: 方向 (-1, 0, 1)
        total: 总长度
    Returns:
        start1, end1, start2, end2: 边界索引
    """
    if direction == 1:
        return 1, total, 0, total - 1
    elif direction == -1:
        return 0, total - 1, 1, total
    else:
        return 0, total, 0, total


class FlowD3(Flow):
    """3D流场类

    实现了3D流场的基本功能,包括:
    - 流场初始化
    - 边界条件处理
    - 流场演化计算

    Args:
        config: 流场配置对象
    """

    def __init__(self, config: FlowConfig, *args, **kwargs):
        super().__init__(config=config, *args, **kwargs)

    def init(self, *args, **kwargs):
        """初始化3D流场

        初始化网格坐标和物理量,包括:
        - 坐标场
        - 速度场
        - 密度场
        - 分布函数等
        """
        m, n, depth = self.config.size
        m, n, depth = int(m + 1), int(n + 1), int(depth + 1)

        # 使用 arange 和 meshgrid 替代循环来初始化坐标
        x = torch.arange(m, device=self.device)
        y = torch.arange(n, device=self.device)
        z = torch.arange(depth, device=self.device)
        X, Y, Z = torch.meshgrid(x, y, z, indexing="ij")

        self.x = torch.stack([X, Y, Z], dim=-1)

        # 初始化物理量
        shape = (m, n, depth)
        self.u = torch.zeros([*shape, 3], device=self.device)
        self.rou = torch.ones([*shape, 1], device=self.device)
        self.tau = torch.full_like(self.rou, 3 * self.config.mu + 0.5)
        self.FOL = torch.zeros_like(self.u)
        self.p = torch.zeros_like(self.rou)
        self.f = torch.zeros([*shape, self.param.e_dim], device=self.device)
        self.f = torch.matmul(self.rou, self.param.w)
        self.update_u_rou()

    @run_timer
    def cul_equ2(self):
        # 预计算一些常量
        cs2 = self.param.cs**2
        cs4 = cs2**2

        for alpha in range(len(self.param.e)):
            t1 = (self.param.e[alpha] - self.u) / cs2
            t2 = ((self.u * self.param.e[alpha]) * self.param.e[alpha]) / cs4
            t3 = torch.sum((t1 + t2) * self.FOL, dim=-1, keepdim=True)
            t4 = (1 - 1 / (2 * self.tau)) * self.param.w[0][alpha] * t3
            self.f[:, :, :, alpha] += t4[:, :, :, 0]

    @run_timer
    def update_u_rou(self, step=0, *args, **kwargs):
        self.rou = torch.sum(self.f, dim=-1, keepdim=True)

        self.p = self.rou / 3.0
        self.u = (torch.matmul(self.f, self.param.e) + self.FOL / 2.0) / self.rou

        # TODO 计算gama
        self.update_u_rou_boundary()

    @run_timer
    def update_u_rou_boundary(self, *args, **kwargs):
        self.u[0, :, :, :] = 0
        self.u[-1, :, :, :] = 0
        self.u[:, 0, :, :] = 0
        self.u[:, -1, :, :] = 0
        self.u[:, :, 0, :] = 0
        self.u[:, :, -1, :] = 0

        if self.config.boundary.input.is_condition(BoundaryCondition.NON_EQUILIBRIUM):
            shape = self.u.shape
            uw = (
                self.config.Re
                * self.config.mu
                / self.rou.max()
                / min(shape[1], shape[2])
            )
            self.u[0, :, :, 0] = init_u(shape[1], shape[2], u_max=uw)
            self.rou[0, :, :, :] = self.rou[1, :, :, :]

        if self.config.boundary.output.is_condition(BoundaryCondition.NON_EQUILIBRIUM):
            self.rou[-1, :, :, :] = self.rou[-2, :, :, :]
            self.u[-1, :, :, :] = self.u[-2, :, :, :]

    @run_timer
    def cul_equ(self, step=0, *args, **kwargs):
        tmp = torch.matmul(self.u, self.param.eT()) / (self.param.cs**2)
        u2 = (
            torch.linalg.norm(self.u, dim=-1, keepdim=True) ** 2
            / (self.param.cs**2)
            / 2
        )
        self.feq = (1 + tmp + tmp**2 / 2 - u2) * torch.matmul(self.rou, self.param.w)
        self.f = self.f - (self.f - self.feq) / self.tau

    @run_timer
    def f_stream(self):
        fcopy = self.f.clone()

        # 跳过静止粒子(e1=e2=e3=0)
        for k, e in enumerate(self.param.e[1:], 1):  # 从索引1开始
            e1, e2, e3 = e

            # 使用列表推导式简化代码
            shifts_dims = [
                (int(shift), dim)
                for shift, dim in zip([e1, e2, e3], range(3))
                if shift != 0
            ]

            f_temp = fcopy[..., k]
            for shift, dim in shifts_dims:
                f_temp = torch.roll(f_temp, shifts=shift, dims=dim)
            self.f[..., k] = f_temp

        self.f_stream_bound(fcopy)

    @run_timer
    def f_stream_bound(self, fcopy):
        bound_config = self.config.boundary

        # 注意：入口/出口边界的处理必须在其他边界之前，
        # 因为入口/出口边界的处理可能会影响到其他边界
        self.f_stream_bound_input(fcopy, bound_config.input)
        self.f_stream_bound_output(fcopy, bound_config.output)

        # 其他边界的处理顺序不重要
        self.f_stream_bound_y_start(fcopy, bound_config.back)
        self.f_stream_bound_y_end(fcopy, bound_config.forward)
        self.f_stream_bound_z_start(fcopy, bound_config.bottom)
        self.f_stream_bound_z_end(fcopy, bound_config.top)

    def f_stream_bound_input(self, fcopy: torch.Tensor, boundary: Boundary) -> None:
        index = self.param.vertex_index(0, 1)
        if boundary.is_condition(BoundaryCondition.PERIODICAL):
            self.f[:1, :, :, index] = fcopy[-1:, :, :, index]
        elif boundary.is_condition(BoundaryCondition.WALL):
            self.f[:1, :, :, index] = fcopy[:1, :, :, self.param.index_reverse(index)]
        elif boundary.is_condition(BoundaryCondition.WALL_WITH_SPEED):
            tmp = 2 * torch.matmul(self.rou[:1, :, :, :], self.param.w[:, index])
            tmp = (
                tmp
                * torch.matmul(self.param.e[index], boundary.get("uw"))
                / self.param.cs**2
            )
            self.f[:1, :, :, index] = (
                fcopy[:1, :, :, self.param.index_reverse(index)] - tmp
            )
        elif boundary.is_condition(BoundaryCondition.NON_EQUILIBRIUM):
            self.f[:1, :, :, :] = self.feq[:1, :, :, :] + (
                self.f[1:2, :, :, :] - self.feq[1:2, :, :, :]
            )
        else:
            raise NotImplementedError

    def f_stream_bound_output(self, fcopy: torch.Tensor, boundary: Boundary) -> None:
        # 右边
        index = self.param.vertex_index(0, -1)
        if boundary.is_condition(BoundaryCondition.PERIODICAL):
            self.f[-1:, :, :, index] = fcopy[:1, :, :, index]
        elif boundary.is_condition(BoundaryCondition.WALL):
            self.f[-1:, :, :, index] = fcopy[-1:, :, :, self.param.index_reverse(index)]
        elif boundary.is_condition(BoundaryCondition.WALL_WITH_SPEED):
            tmp = 2 * torch.matmul(self.rou[-1:, :, :, :], self.param.w[:, index])
            tmp = (
                tmp
                * torch.matmul(self.param.e[index], boundary.get("uw"))
                / self.param.cs**2
            )
            self.f[-1:, :, :, index] = (
                fcopy[-1:, :, :, self.param.index_reverse(index)] - tmp
            )
        elif boundary.is_condition(BoundaryCondition.NON_EQUILIBRIUM):
            self.f[-1:, :, :, :] = self.feq[-1:, :, :, :] + (
                self.f[-2:-1, :, :, :] - self.feq[-2:-1, :, :, :]
            )
        else:
            raise NotImplementedError

    def f_stream_bound_y_start(self, fcopy, boundary: Boundary):
        # 后面
        index = self.param.vertex_index(1, 1)
        if boundary.is_condition(BoundaryCondition.PERIODICAL):
            self.f[:, :1, :, index] = fcopy[:, -1:, :, index]
        elif boundary.is_condition(BoundaryCondition.WALL):
            self.f[:, :1, :, index] = fcopy[:, :1, :, self.param.index_reverse(index)]
        elif boundary.is_condition(BoundaryCondition.WALL_WITH_SPEED):
            tmp = 2 * torch.matmul(self.rou[:, :1, :, :], self.param.w[:, index])
            tmp = (
                tmp
                * torch.matmul(self.param.e[index], boundary.get("uw"))
                / self.param.cs**2
            )
            self.f[:, :1, :, index] = (
                fcopy[:, :1, :, self.param.index_reverse(index)] - tmp
            )
        else:
            raise NotImplementedError

    def f_stream_bound_y_end(self, fcopy, boundary: Boundary):
        # 前面
        index = self.param.vertex_index(1, -1)
        if boundary.is_condition(BoundaryCondition.PERIODICAL):
            self.f[:, -1:, :, index] = fcopy[:, :1, :, index]
        elif boundary.is_condition(BoundaryCondition.WALL):
            self.f[:, -1:, :, index] = fcopy[:, -1:, :, self.param.index_reverse(index)]
        elif boundary.is_condition(BoundaryCondition.WALL_WITH_SPEED):
            tmp = 2 * torch.matmul(self.rou[:, -1:, :, :], self.param.w[:, index])
            tmp = (
                tmp
                * torch.matmul(self.param.e[index], boundary.get("uw"))
                / self.param.cs**2
            )
            self.f[:, -1:, :, index] = (
                fcopy[:, -1:, :, self.param.index_reverse(index)] - tmp
            )
        else:
            raise NotImplementedError

    def f_stream_bound_z_start(self, fcopy, boundary: Boundary):
        # 下面
        index = self.param.vertex_index(2, 1)
        if boundary.is_condition(BoundaryCondition.PERIODICAL):
            self.f[:, :, :1, index] = fcopy[:, :, -1:, index]
        elif boundary.is_condition(BoundaryCondition.WALL):
            self.f[:, :, :1, index] = fcopy[:, :, :1, self.param.index_reverse(index)]
        elif boundary.is_condition(BoundaryCondition.WALL_WITH_SPEED):
            tmp = 2 * torch.matmul(self.rou[:, :, :1, :], self.param.w[:, index])
            tmp = (
                tmp
                * torch.matmul(self.param.e[index], boundary.get("uw"))
                / self.param.cs**2
            )
            self.f[:, :, :1, index] = (
                fcopy[:, :, :1, self.param.index_reverse(index)] - tmp
            )
        else:
            raise NotImplementedError

    def f_stream_bound_z_end(self, fcopy, boundary: Boundary):
        # 上面
        index = self.param.vertex_index(2, -1)
        if boundary.is_condition(BoundaryCondition.PERIODICAL):
            self.f[:, :, -1:, index] = fcopy[:, :, :1, index]
        elif boundary.is_condition(BoundaryCondition.WALL):
            self.f[:, :, -1:, index] = fcopy[:, :, -1:, self.param.index_reverse(index)]
        elif boundary.is_condition(BoundaryCondition.WALL_WITH_SPEED):
            tmp = 2 * torch.matmul(self.rou[:, :, -1:, :], self.param.w[:, index])
            tmp = (
                tmp
                * torch.matmul(self.param.e[index], boundary.get("uw"))
                / self.param.cs**2
            )
            self.f[:, :, -1:, index] = (
                fcopy[:, :, -1:, self.param.index_reverse(index)] - tmp
            )
        else:
            raise NotImplementedError



from .base import Flow, FlowConfig
from .d3 import FlowD3, FlowD3Q13, FlowD3Q15, FlowD3Q19, FlowD3Q27

__all__ = [
    "Flow",
    "FlowD3",
    "FlowD3Q19",
    "FlowD3Q13",
    "FlowConfig",
    "create_flow",
    "FlowD3Q15",
    "FlowD3Q27",
]


def create_flow(flow_config: FlowConfig, *args, **kwargs):
    # return FlowD3(config=flow_config, *args, **kwargs)

    """解析3D模型参数

    根据参数类型创建对应的参数对象

    Args:
        flow_type: 参数类型,支持"D3Q27"/"D3Q19"/"D3Q15"/"D3Q13"

    Returns:
        Param: 参数对象

    Raises:
        ValueError: 当参数类型不支持时
    """
    if flow_config.param_type == "D3Q27":
        return FlowD3Q27(config=flow_config, *args, **kwargs)
    elif flow_config.param_type == "D3Q19":
        return FlowD3Q19(config=flow_config, *args, **kwargs)
    elif flow_config.param_type == "D3Q15":
        return FlowD3Q15(config=flow_config, *args, **kwargs)
    elif flow_config.param_type == "D3Q13":
        return FlowD3Q13(config=flow_config, *args, **kwargs)
    else:
        raise ValueError("Unknown parameter type: {}".format(flow_config.param_type))



import h5py
import torch
from funutil import deep_get, run_timer

from funlbm.base import Worker
from funlbm.config.base import BaseConfig
from funlbm.particle.coord import CoordConfig, Coordinate
from funlbm.util import tensor_format


class ParticleConfig(BaseConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.coord_config: CoordConfig = CoordConfig()
        self.type = "ellipsoid"

    def _from_json(self, config_json: dict, *args, **kwargs):
        self.coord_config.from_json(deep_get(config_json, "coord"))
        self.type = deep_get(config_json, "type") or self.type


class Particle(Worker):
    """
    粒子基类

    属性:
        config (ParticleConfig): 粒子配置
        coord (Coordinate): 坐标系对象
        mass (float): 粒子质量
        area (float): 粒子表面积
        I (Tensor): 惯性矩
        rou (float): 粒子密度
        angle (Tensor): 粒子方向
        cx (Tensor): 质心坐标 [x,y,z]
        cr (Tensor): 质心半径 [a,b,b]
        cu (Tensor): 质心速度 [vx,vy,vz]
        cw (Tensor): 质心角速度 [wx,wy,wz]
        cF (Tensor): 质心合外力
        cT (Tensor): 质心合外力矩
        lx (Tensor): 拉格朗日点坐标 [m,i,3]
        lF (Tensor): 拉格朗日点受力 [m,i,3]
        lm (Tensor): 拉格朗日点质量
        lu (Tensor): 拉格朗日点速度 [m,i,3]
        lrou (Tensor): 拉格朗日点密度
    """

    def __init__(self, config: ParticleConfig = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.config: ParticleConfig = config or ParticleConfig()
        self.coord: Coordinate = Coordinate(
            config=self.config.coord_config, *args, **kwargs
        )

        # 颗粒质量[1]
        self.mass = None
        # 颗粒面积[1]
        self.area = None
        # 惯性矩
        self.I = None
        # 颗粒密度[1]
        self.rou = None
        # 颗粒方向[i,j,k]
        self.angle = None

        # 质心坐标[i,j,k]
        self.cx = torch.tensor(
            self.config.coord_config.center, device=self.device, dtype=torch.float32
        )
        # 质心半径[a,b,b]
        self.cr = 5 * torch.ones(5, device=self.device, dtype=torch.float32)
        # 质心速度[i,j,k]
        self.cu = torch.zeros(3, device=self.device, dtype=torch.float32)
        # 质心角速度[i,j,k]
        self.cw = torch.zeros(3, device=self.device, dtype=torch.float32)
        # 质心合外力
        self.cF = torch.zeros(3, device=self.device, dtype=torch.float32)
        # 质心合外力
        self.cT = torch.zeros(3, device=self.device, dtype=torch.float32)

        self._lagrange: torch.Tensor = torch.zeros(
            [0], device=self.device, dtype=torch.float32
        )
        # 拉格朗日点的坐标[m,i,3]
        self.lx: torch.Tensor = torch.zeros(
            [0], device=self.device, dtype=torch.float32
        )
        # 拉格朗日点上的力[m,i,3]
        self.lF: torch.Tensor = torch.zeros(
            [0], device=self.device, dtype=torch.float32
        )
        # 拉格朗日点的质量
        self.lm: torch.Tensor = torch.zeros(
            [0], device=self.device, dtype=torch.float32
        )
        # 拉格朗日点速度[m,i,3]
        self.lu: torch.Tensor = torch.zeros(
            [0], device=self.device, dtype=torch.float32
        )
        # 拉格朗日点速度[m,i,3]
        self.lrou: torch.Tensor = torch.zeros(
            [0], device=self.device, dtype=torch.float32
        )

    def _init(self, dx=1, *args, **kwargs):
        raise NotImplementedError("还没实现")

    def init(self, *args, **kwargs):
        self.rou = float(self.config.get("rou", 1.0))
        self._init(*args, **kwargs)
        shape = self._lagrange.shape
        self.lx = torch.zeros_like(self._lagrange, dtype=torch.float32)
        self.lF = torch.zeros_like(self._lagrange, dtype=torch.float32)
        self.lu = torch.zeros_like(self._lagrange, dtype=torch.float32)
        self.lm = torch.full(
            (shape[0], 1), self.area / shape[0], device=self.device, dtype=torch.float32
        )
        self.lrou = torch.empty((shape[0], 1), device=self.device, dtype=torch.float32)

    @run_timer
    def update_from_lar(self, dt: float, gl: float = 9.8, rouf: float = 1.0) -> None:
        """
        从拉格朗日坐标更新粒子状态。

        参数:
            dt: 时间步长
            gl: 重力加速度
            rouf: 流体密度

        异常:
            ValueError: 当输入参数无效时抛出
        """
        if dt <= 0:
            raise ValueError("Time step must be positive")
        if rouf <= 0:
            raise ValueError("Fluid density must be positive")

        tmp = (
            (1 - self.rou / rouf)
            * self.mass
            * torch.tensor([gl, 0, 0], device=self.device)
        )
        self.cF = torch.sum(-self.lF * self.lm, dim=0) + tmp

        self.cu = self.cu + self.cF / self.mass * dt
        self.cx = self.cx + self.cu * dt

        self.cT = -torch.sum(
            torch.cross(self.lx - self.cx, self.lF, dim=-1) * self.lm, dim=0
        )
        self.cw = self.cw + 0.1 * self.cT * dt / self.I

    @run_timer
    def update(self, dt):
        self.coord.update(center=self.cx, w=self.cw)
        self.lx = self.coord.cul_point(self._lagrange)

    def from_json(self):
        pass

    def to_str(self, step=0, *args, **kwargs):
        res = f"m={self.mass:.2f}"
        res += "\tcu=" + ",".join([f"{i:.6f}" for i in self.cu])
        res += "\tcx=" + ",".join([f"{i:.6f}" for i in self.cx])
        res += "\tcf=" + ",".join([f"{i:.6f}" for i in self.cF])
        res += "\tlF=" + ",".join(
            [f"{i:.6f}" for i in [self.lF.min(), self.lF.mean(), self.lF.max()]]
        )
        res += "\tcw=" + ",".join([f"{i:.6f}" for i in self.cw])
        res += "\tr=" + self.coord.to_str()
        return res

    def to_json(self, step=0, *args, **kwargs):
        return {
            "m": float(self.mass.cpu().numpy()),
            "cu": tensor_format(self.cu),
            "cx": tensor_format(self.cx),
            "cf": tensor_format(self.cF),
            "lF": tensor_format(
                [
                    self.lF.min(),
                    self.lF.mean(),
                    self.lF.max(),
                ]
            ),
            "cw": tensor_format(self.cw),
            "coord": self.coord.to_json(),
        }

    def dump_checkpoint(self, group: h5py.Group = None, all=False, *args, **kwargs):
        if group is None:
            return

        self.dupmp_dataset(group, "cu", self.cu)
        self.dupmp_dataset(group, "cx", self.cx)
        self.dupmp_dataset(group, "cF", self.cF)
        self.dupmp_dataset(group, "cw", self.cw)

        self.dupmp_dataset(group, "lx", self.lx)
        self.dupmp_dataset(group, "lF", self.lF)
        self.dupmp_dataset(group, "lu", self.lu)
        self.dupmp_dataset(group, "lrou", self.lrou)

    def load_checkpoint(self, group: h5py.Group = None, *args, **kwargs):
        if group is None:
            return
        self.cu = self.load_dataset(group, "cu")
        self.cx = self.load_dataset(group, "cx")
        self.cF = self.load_dataset(group, "cF")
        self.cw = self.load_dataset(group, "cw")
        self.lx = self.load_dataset(group, "lx")
        self.lF = self.load_dataset(group, "lF")
        self.lu = self.load_dataset(group, "lu")
        self.lrou = self.load_dataset(group, "lrou")
